@@ -7,7 +7,8 @@ from os import listdir, makedirs
 from os.path import isdir, join
 from osgeo.gdal import Open
 import tensorflow as tf
-from tensorflow.keras.layers import Add, BatchNormalization, Concatenate, Conv2D, Dropout, Input, MaxPooling2D, UpSampling2D
+from tensorflow.keras.layers import Add, BatchNormalization, Concatenate, Conv2D, Dropout, Input, MaxPooling2D, \
+    UpSampling2D
 from tensorflow.keras.layers.experimental.preprocessing import Rescaling
 
 
@@ -15,8 +16,6 @@ class SegmentationModel:
 
     def __init__(self):
 
-        super().__init__()
-
         self.test_metrics = {}
         self.test_filenames = []
         self.model = None
@@ -59,7 +58,7 @@ class SegmentationModel:
             labels = unique(y_true)
 
             # compute total accuracy
-            metrics_dict["total_accuracy"] = total_accuracy(y_true, y_pred)
+            metrics_dict['total_accuracy'] = total_accuracy(y_true, y_pred)
 
             # compute F1 and Jaccard scores for each label
             f1_scores = []
@@ -74,8 +73,8 @@ class SegmentationModel:
                                               pos_label=label))
 
             # add F1 and Jaccard scores to the metrics dictionary
-            metrics_dict["F1"] = f1_scores
-            metrics_dict["Jaccard"] = jaccard_scores
+            metrics_dict['F1'] = f1_scores
+            metrics_dict['Jaccard'] = jaccard_scores
 
             fname_metrics[fname] = metrics_dict
 
@@ -151,7 +150,9 @@ class VGG19Unet(SegmentationModel):
     def __init__(self, n_channels: int = 3,
                  n_classes: int = 2,
                  n_filters: int = 16,
-                 dropout_rate: float = 0.3):
+                 dropout_rate: float = 0.3,
+                 rescale_factor: float = 1 / 255,
+                 include_residual: bool = False):
 
         # initialize the superclass
         super().__init__()
@@ -162,35 +163,13 @@ class VGG19Unet(SegmentationModel):
         self.n_filters = n_filters
         self.dropout_rate = dropout_rate
 
-        if n_channels == 1:
-            self.activation = 'sigmoid'
-        else:
-            self.activation = 'softmax'
-
-    def compile_model(self, loss: tf.keras.losses.Loss = 'sparse_categorical_crossentropy',
-                      learning_rate: float = 0.001,
-                      rescale_factor: float = 1 / 255,
-                      include_residual: bool = False,
-                      include_attention: bool = False
-                      ) -> None:
-
-        """Returns a model object, compiled with the provided loss and optimizer. Additionally, this sets the self.model
-        attribute with the compiled model.
-
-        Args:
-            loss: the loss function to use during training;
-            learning_rate: the starting learning rate for the Adam optimizer;
-            rescale_factor: the factor by which to rescale the input tensor;
-            include_residual: incorporate a residual connection into each convolutional block;
-            include_attention: incorporate an attention module into each skip-connection.
-
-        Returns:
-            None"""
+        # ensure that n_classes >= 2
+        if self.n_classes < 2:
+            raise Exception("Number of classes must at least 2.")
 
+        # define the layers and model
         include_dropout = (self.dropout_rate > 0.0)
 
-        # build the model graph
-
         def conv_block(input_tensor,
                        filters):
 
@@ -203,8 +182,6 @@ class VGG19Unet(SegmentationModel):
 
             return batch_norm
 
-        # build the model graph
-
         # level 0
         inputs = Input(shape=(None, None, self.n_channels), dtype=tf.float32)
         d0 = Rescaling(scale=rescale_factor)(inputs)
@@ -301,17 +278,28 @@ class VGG19Unet(SegmentationModel):
         outputs = Conv2D(filters=self.n_classes,
                          kernel_size=(1, 1),
                          padding='same',
-                         activation=self.activation)(u0_out)
+                         activation='softmax')(u0_out)
 
         # create the model object
-        model = tf.keras.Model(inputs=inputs, outputs=outputs)
+        self.model = tf.keras.Model(inputs=inputs, outputs=outputs)
+
+    def compile(self, loss=None,
+                learning_rate: float = 0.001) -> None:
+
+        """Compiles the model attribute with the given loss function, and an Adam optimizer with the provided learning
+        rate.
+
+        Args:
+            loss: the loss function to use during training;
+            learning_rate: the starting learning rate for the Adam optimizer.
+
+        Returns:
+            None
+        """
 
         # compile the model
-        model.compile(loss=loss,
-                      optimizer=tf.keras.optimizers.Adam(learning_rate=learning_rate,
-                                                         clipnorm=1.0))
-
-        self.model = model
+        self.model.compile(loss=loss,
+                           optimizer=tf.keras.optimizers.Adam(learning_rate=learning_rate))
 
 
 class Unet(SegmentationModel):
@@ -319,7 +307,9 @@ class Unet(SegmentationModel):
     def __init__(self, n_channels: int = 3,
                  n_classes: int = 2,
                  n_filters: int = 16,
-                 dropout_rate: float = 0.3):
+                 dropout_rate: float = 0.3,
+                 rescale_factor: float = 1 / 255,
+                 include_residual: bool = False):
 
         # initialize the superclass
         super().__init__()
@@ -330,29 +320,9 @@ class Unet(SegmentationModel):
         self.n_filters = n_filters
         self.dropout_rate = dropout_rate
 
-        if n_channels == 1:
-            self.activation = 'sigmoid'
-        else:
-            self.activation = 'softmax'
-
-    def compile_model(self, loss: tf.keras.losses.Loss = 'sparse_categorical_crossentropy',
-                      learning_rate: float = 0.001,
-                      rescale_factor: float = 1 / 255,
-                      include_residual: bool = False,
-                      include_attention: bool = False) -> None:
-
-        """Returns a model object, compiled with the provided loss and optimizer. Additionally, this sets the self.model
-        attribute with the compiled model.
-
-        Args:
-            loss: the loss function to use during training;
-            learning_rate: the starting learning rate for the Adam optimizer;
-            rescale_factor: the factor by which to rescale the input tensor;
-            include_residual: include residual connections in each level;
-            include_attention: include an attention module for each skip-connection.
-
-        Returns:
-            None"""
+        # ensure that n_classes >= 2
+        if self.n_classes < 2:
+            raise Exception("Number of classes must at least 2.")
 
         include_dropout = (self.dropout_rate > 0.0)
 
@@ -438,13 +408,25 @@ class Unet(SegmentationModel):
         outputs = Conv2D(filters=self.n_classes,
                          kernel_size=(1, 1),
                          padding='same',
-                         activation=self.activation)(u0_out)
+                         activation='softmax')(u0_out)
 
         # create the model object
-        model = tf.keras.Model(inputs=inputs, outputs=outputs)
+        self.model = tf.keras.Model(inputs=inputs, outputs=outputs)
+
+    def compile(self, loss=None,
+                learning_rate: float = 0.001) -> None:
+
+        """Compiles the model attribute with the given loss function, and an Adam optimizer with the provided learning
+        rate.
+
+        Args:
+            loss: the loss function to use during training;
+            learning_rate: the starting learning rate for the Adam optimizer.
+
+        Returns:
+            None
+        """
 
         # compile the model
-        model.compile(loss=loss,
-                      optimizer=tf.keras.optimizers.Adam(learning_rate=learning_rate))
-
-        self.model = model
+        self.model.compile(loss=loss,
+                           optimizer=tf.keras.optimizers.Adam(learning_rate=learning_rate))
